function [avg_timings, avg_stats, all_extra_info] = ...
    rigor_analysis(segobj_filepaths, output_dir, per_im_out, catg_strs, ...
                   force_timing_scale)
% This function generates average statistics (including quantitative
% scores) for the proposals generated by RIGOR. Some part of the function
% is involved in computing the avg scores for all the images in the set 
% (to get this, you need to run it after running compute_print_results on
% each image). This would compute the average segmentation overlap score,
% covering, average bounding box overlap score, etc. If atleast one image
% in the set has scores, this function would generate scores.eps image
% summarizing scores for images in different ways.
%
% avg_stats, all_extra_info: To undertand the content of these structs, we
% should know that RIGOR performs parametric min-cut over different graph 
% types, each of which specifies unaries differently. The two major graph 
% types are uniform and color graphs (see code). Uniform graphs define 
% unaries based on the size of the superpixel. Color graphs define unaries 
% based on the color difference to the seed superpixels. Each major graph 
% type has multiple sub-types - each defining unaries slightly differently. 
% RIGOR has 3 different types of uniform graphs and 3 different color 
% graphs. In total RIGOR (CVPR'14) has G=6 graph-types.
%
% <all_extra_info>: stores quantitative scores about all N ground-truths 
%  |            (over I images):
%  |-- <im_no> is a 1xN vector giving the image index for each GT.
%  |-- <gt_category_id> is a 1xN vector which gives the category id for 
%  |            each GT.
%  |-- <best_overlaps> is a 1xN vector of best overlaps for each GT. The 
%  |            best overlapping segment is selected from the set of 
%  |            proposals generated for image im_no(i).
%  |-- <best_ov_seg_idx> is a 1xN vector giving the best overlap proposal 
%  |            index for each GT i.e. this proposal, for image <im_no>(i), 
%  |            gives the best overlap score for the current GT.
%  |-- <best_overlaps_sgraphs> is a NxG matrix, where each column gives 
%  |            best overlaps for one graph sub-type. 
%  |            max(best_overlaps_sgraphs,[],2) == best_overlaps'
%  |-- <sgraph_gt_wins> is a IxG matrix, where each column gives the number 
%  |            of times a graph sub-type gives the best overlap (the 
%  |            winning proposal). sum(sgraph_gt_wins,2) is the number of 
%  |            GTs in each image.
%  |-- <gt_szs> is a 1xN vector giving the size in pixels for each GT.
%  |-- <bb_overlaps> is similar to <best_overlaps>, except that it is over 
%  |            bounding boxes (of a GT segment) rather than the GT 
%  |            segmentation.
%  |-- <best_bbov_seg_idx> is the bounding box counterpart of 
%  |            <best_ov_seg_idx>.
%  |-- <bb_szs> is the bounding box counterpart of <gt_szs> (size of the 
%  |            bounding box rectangle).
%  |-- <sgraph_names> is a 1xG cell array, where each cell is a string 
%  |            identifying a graph sub-type.
%
% <avg_stats>: stores average quantitative scores and some other 
%  |            information:
%  |-- <final_num_segs> Gives the average number of proposals generated 
%  |            over I images in the dataset.
%  |-- <avg_best_overlap> Is the average overlap over all GTs. 
%  |            = mean(all_extra_info.best_overlaps)
%  |-- <covering> Is the covering computed by averaging the covering scores 
%  |            for all I images.
%  |-- <collated_covering> Covering computed by considering all GTs 
%  |            together i.e.
%  |            = sum(all_extra_info.best_overlaps.*all_extra_info.gt_szs) 
%  |              / sum(all_extra_info.gt_szs);
%  |-- <avg_best_overlaps_sgraphs> Is a 1xG vector average overlap scores 
%  |            over each graph sub-type, i.e. 
%  |            = mean(all_extra_info.best_overlaps_sgraphs)
%  |-- <all_sgraph_gt_wins> Is a 1xG vector giving the number of times each 
%  |            graph sub-type produces a proposal with overall best 
%  |            overlap, i.e. = sum(all_extra_info.sgraph_gt_wins). 
%  |            Also sum(all_sgraph_gt_wins) = N
%  |-- <avg_bb_overlap> The bounding box counterpart of <avg_best_overlap>, 
%  |            i.e. = mean(all_extra_info.bb_overlaps)
%  |-- <num_gts> = N
%  |-- <num_ims> = I
%  |-- <gt_categories> Is a 1xC cell array, which gives names of each of 
%  |            the C GT categories. The ordering is according to the ids 
%  |            in <gt_category_id>
%  |-- <category_avg_overlaps: Is a 1xC array, which gives the average 
%  |            best overlap for each GT category i.e. 
%  |            = accumarray(all_extra_info.gt_category_id(:), 
%  |                         all_extra_info.best_overlaps, [], @mean)
%  |-- <num_category_gts> Is the number of instances for each GT category, 
%  |            i.e. = accumarray(all_extra_info.gt_category_id(:), 1)
%  |-- <seg_recall> Is a 1x3 array giving recall/detection rates at 0.5, 
%  |            0.7, 0.9 overlaps, i.e. percentage of GTs which have an 
%  |            overlap above 0.5, 0.7, or 0.9. For instance at 0.5 the 
%  |            recall is computed as 
%  |            nnz(all_extra_info.best_overlaps>=0.5) / avg_stats.num_gts
%  |-- <bb_recall> Is the bounding box counterpart of <seg_recall>.
%
% This function also analyzes per image timing of each component of
% rigor_obj_segments. It breaks it down into the components in the pipeline
% (e.g. max-flow) and its sub-components (e.g. actual max-flow and its
% overheads). It generates a pipeline timing image (if output_dir is 
% provided and is not empty) for each image passed in segobj_filepaths. 
% segobj_filepaths can be a list of files or just a directory path 
% containing mat files saved by rigor_obj_segments. It also generates an 
% average image (avg_timing.eps) and a summary average image which collates
% times for sub-components (avg_timing_summary.eps).
%
% The timing break-down
%   - Everything under <total_init_time> is a single time. Everything under
%     <total_computing_segs_time> are given individually for each graph
%     type. Hence, the length of the array for each field there is equal to
%     the number of graph types used.
%
%   |-- <total_seg_time>
%       |-- Setting of the segmenter
%       |-- <total_init_time> Initialize segmenter (Segmenter.initialize)
%       |   |-- <extra_bndry_compute_time> If the Pb file is loaded rather
%       |   |               than computed, this time non-zero indicating 
%       |   |               the time it took to originally compute it. 
%       |   |               NOTE: this is would be extra time actually not 
%       |   |               accounted for in the total time.
%       |   |-- <im_pairwise_time> Time for loading Pb. So the actual time
%       |   |               for computing Pb would be <im_pairwise_time> + 
%       |   |               <extra_bndry_compute_time>
%       |   |-- <superpixels_compute_time> Time taken to compute 
%       |   |               superpixels
%       |   |-- <precompute_pairwise_time> Time taken to precompute the
%       |   |               pairwise edge potentials i.e. precompute edge 
%       |   |               capacities
%       |   |-- <precompute_seed_time> Time taken to precompute the seed
%       |   |               locations and any other seed computation
%       |-- <total_computing_segs_time> Time to actually create the
%       |   |           segments (Segmenter.compute_segments)
%       |   |-- <unary_cost_set_time> Time to compute all the different
%       |   |               unary costs used for the parametric min-cut
%       |   |-- <pairwise_set_time> Time taken for setting the pairwise
%       |   |               costs in the graph (after computing Pb etc.)
%       |   |-- <pmc_time> Time to do the parametric min cut for all the
%       |   |   |           graphs created
%       |   |   |-- <pmc_parallel_cut_time> The time it takes to actually
%       |   |   |           do the cut. Note, this is done in parallel so
%       |   |   |           this time might be greater than <pmc_time>
%       |   |   |-- <pmc_parallel_overhead_time> The time it takes to set
%       |   |   |           up the graph and its values for parametric min
%       |   |   |           cut. Note, this is done in parallel so this
%       |   |   |           time might be greater than <pmc_time>
%       |   |-- <seg_filtering_time> Time to filter segments
%       |   |   |-- <init_filter_time> Time taken to initialize the segment
%       |   |   |           filteration (which includes the time to 
%       |   |   |           separate connected components)
%       |   |   |-- <energy_filter_time> Time to compute the energy for 
%       |   |   |           each segment (cut ratio), and then filtering
%       |   |   |           based on max energy
%       |   |   |-- <rand_filter_time> Time taken to randomly select
%       |   |   |           segments if more than a certain amount (based 
%       |   |   |           on segm_pars.randomize_N)
%       |   |   |-- <seg_similar_filter_time> Filtering similar segments
%       |   |   |           (this includes removing duplicate segments and
%       |   |   |           filteration based on minimum dissimilarity)
%
%
% @authors:     Ahmad Humayun
% @contact:     ahumayun@cc.gatech.edu
% @affiliation: Georgia Institute of Technology
% @date:        Fall 2013 - Fall 2014

    dFig('addpath_export_fig');
    
    % RIGOR to the path
    rigor_root = fullfile(fileparts(which(mfilename)), '..');
    addpath(rigor_root);
    addpath(fullfile(rigor_root, 'utils/'));
    
    if ~exist('segobj_filepaths', 'var') || isempty(segobj_filepaths)
        segobj_filepaths = {};
        sub_dir = 'data/MySegmentsMat';
        
        segs_dir = fullfile(rigor_root, sub_dir);
        d = dir(fullfile(segs_dir));
        for i = 1:length(d)
            if d(i).name(1) ~= '.' && d(i).isdir
                f = dir(fullfile(segs_dir, d(i).name, '*_segobj.mat'));
                segobj_filepaths = ...
                    arrayfun(@(x) fullfile(segs_dir, d(i).name, x.name), ...
                             f, 'UniformOutput',false);
                if ~isempty(segobj_filepaths)
                    break;
                end
            end
        end
        
        assert(~isempty(segobj_filepaths), 'Cannot find any mat files');
    elseif ischar(segobj_filepaths)
        % incase a directory path was provided rather than a list of files
        f = dir(fullfile(segobj_filepaths, '*_segobj.mat'));
        segobj_filepaths = ...
            arrayfun(@(x) fullfile(segobj_filepaths, x.name), ...
                     f, 'UniformOutput',false);
    elseif isstruct(segobj_filepaths)
        fprintf('Computing file ordering\n');
        
        base_path = segobj_filepaths.base_path;
        batch_files = importdata(segobj_filepaths.batchlist_filepath);
        
        % get the full saved filepath for each image
        full_fp = cellfun(@(x) fullfile(base_path,[x,'_segobj.mat']), ...
                              batch_files, 'UniformOutput',false);
        % check if each saved file exists
        file_exist = cellfun(@(x) exist(x,'file')==2, full_fp);
        
        % get the files that don't exist
        no_f = batch_files(~file_exist);
        if ~isempty(no_f) > 0
            warning('Couldn''t find results for file %s\n', no_f{:});
            warning('Total %d/%d couldn''t be found', numel(no_f), ...
                    numel(batch_files));
        end
        
        segobj_filepaths = full_fp(file_exist);
    end
    
    if ~exist('output_dir', 'var'), output_dir = []; end
    if ~exist('per_im_out', 'var'), per_im_out = true; end;
    if ~exist('force_timing_scale', 'var'), force_timing_scale = []; end
    if ~exist('catg_strs', 'var'), catg_strs = {}; end
    
    % create directoy if non-existant
    if ~isempty(output_dir) && ~exist(output_dir, 'dir')
        mkdir(output_dir);
    end
    avg_timings = struct;
    avg_stats = struct;
    timings_fields = {};
    extra_info_fields = {};
    % create the set of scoring stats
    all_extra_info.im_no = [];
    all_extra_info.gt_category_id = [];
    all_extra_info.best_overlaps = [];
    all_extra_info.best_ov_seg_idx = [];
    all_extra_info.best_overlaps_sgraphs = [];
    all_extra_info.sgraph_gt_wins = [];
    all_extra_info.gt_szs = [];
    all_extra_info.bb_overlaps = [];
    all_extra_info.best_bbov_seg_idx = [];
    all_extra_info.bb_szs = [];
    all_extra_info.sgraph_names = {};
    
    fi = dFig('init_fig', [420, 560], 'visible','off');
    
    ticker_info = timerTicks('init_ticker');
    
    % collate quantitative scores and timings for all the files
    for f_idx = 1:length(segobj_filepaths)
        filepath = segobj_filepaths{f_idx};
        load_mat = load(filepath);
        
        % get the extra info
        extra_info.final_num_segs = ...
                sum(load_mat.seg_obj.num_segs.after_clustering_FINAL);
        
        % find the number of parametric min-cuts which got to full cuts
        num_sgraphs = sum(cellfun(@numel, ...
                    load_mat.seg_obj.segm_params.graph_sub_methods));
        sgraph_idxs = duplicateElems(1:num_sgraphs, ...
                    load_mat.seg_obj.seeds_per_sgraph);
        tmp = [load_mat.seg_obj.meta_cut_segs_info.extra_cut_info];
        extra_info.frac_all_fg_cut = ...
                    accumarray(sgraph_idxs(:), [tmp.all_fg_seg])' ./ ...
                    double(load_mat.seg_obj.seeds_per_sgraph);
        
        % get the scores data, if it was computed
        scores_fp = load_mat.seg_obj.return_scores_filepath();
        if exist(scores_fp, 'file')
            scores_mat = load(scores_fp);

            extra_info.avg_best_overlap = ...
                scores_mat.collated_scores.avg_best_overlap;
            extra_info.covering = ...
                scores_mat.collated_scores.covering;
            
            all_extra_info.im_no = [all_extra_info.im_no, ...
                                    uint32(repmat(f_idx, ...
                                           [1 numel(scores_mat.Q)]))];
            all_extra_info.gt_category_id = ...
                                  [all_extra_info.gt_category_id, ...
                                   [scores_mat.Q.gt_category_id]];
            all_extra_info.best_overlaps = ...
                                  [all_extra_info.best_overlaps, ...
                                   [scores_mat.Q.best_overlap]];
            all_extra_info.best_ov_seg_idx = ...
                                  [all_extra_info.best_ov_seg_idx, ...
                                   [scores_mat.Q.best_seg_idx]];
            sgraph_ov = scores_mat.collated_scores.best_seg_overlap_sgraphs;
            all_extra_info.best_overlaps_sgraphs = ...
                                  [all_extra_info.best_overlaps_sgraphs; ...
                                   sgraph_ov];
            [~,best_sgraph] = max(sgraph_ov, [], 2);
            curr_wins = accumarray(best_sgraph, 1, [size(sgraph_ov,2),1])';
            all_extra_info.sgraph_gt_wins = ...
                                  [all_extra_info.sgraph_gt_wins; ...
                                   uint32(curr_wins)];
            all_extra_info.gt_szs = [all_extra_info.gt_szs, ...
                                   [scores_mat.Q.gt_seg_szs]];
            temp = [scores_mat.Q.extra_info];
            if isfield(temp, 'best_bb_overlap')
                all_extra_info.bb_overlaps = [all_extra_info.bb_overlaps, ...
                                       [temp.best_bb_overlap]];
                all_extra_info.best_bbov_seg_idx = ...
                                      [all_extra_info.best_bbov_seg_idx, ...
                                       [temp.best_bb_seg_idx]];
                all_extra_info.bb_szs = [all_extra_info.bb_szs, ...
                                       [scores_mat.Q.gt_bb_szs]];
            end
            
            % if subgraph names haven't been stored
            if isempty(all_extra_info.sgraph_names)
                sg_idxs = scores_mat.collated_scores.subgraph_to_sgmntr_idx;
                snames = scores_mat.einfo.sgmntr_names;
                sgnames = cellfun(@(x) x([1:3 end]), ...
                                  scores_mat.einfo.sgraph_names, ...
                                  'UniformOutput',false);
                all_extra_info.sgraph_names = ...
                    arrayfun(@(i) [snames{sg_idxs(i)} ' ' sgnames{i}], ...
                             1:size(sgraph_ov,2), 'UniformOutput',false);
            end
        end
        
        extra_info_fields = fieldnames(extra_info);
        
        % sum the timing across all segmenters
        timings = load_mat.seg_obj.timings;
        timings_fields = fieldnames(timings);
        all_timings = struct;
        for idx = 1:length(timings_fields)
            all_timings.(timings_fields{idx}) = ...
                sum([timings.(timings_fields{idx})]);
        end
        
        % adjust timing if <extra_bndry_compute_time> is non-zero
        if timings.extra_bndry_compute_time > 0
            % if loaded everytime (computation for Pb was done in some 
            % other run)
            all_timings.im_pairwise_time = ...
                all_timings.im_pairwise_time + ...
                all_timings.extra_bndry_compute_time;
            all_timings.total_init_time = ...
                all_timings.total_init_time + ...
                all_timings.extra_bndry_compute_time;
            all_timings.total_seg_time = all_timings.total_seg_time + ...
                all_timings.extra_bndry_compute_time;
        end
        
        % collate the timings and the extra infos
        if f_idx == 1
            avg_timings = all_timings;
            avg_stats = extra_info;
        else
            for idx = 1:length(timings_fields)
                avg_timings.(timings_fields{idx}) = ...
                    avg_timings.(timings_fields{idx}) + ...
                    all_timings.(timings_fields{idx});
            end
            for idx = 1:length(extra_info_fields)
                avg_stats.(extra_info_fields{idx}) = ...
                    [avg_stats.(extra_info_fields{idx}); ...
                     extra_info.(extra_info_fields{idx})];
            end
        end
        
        % draw everything in a plot
        if per_im_out && ~isempty(output_dir)
            plot_timing_bar_chart(all_timings, extra_info, ...
                                  load_mat.seg_obj.input_info.img_name, ...
                                  output_dir, fi);
        end
        
        ticker_info = timerTicks('tick', ticker_info, f_idx, ...
                                 length(segobj_filepaths));
%         fprintf('Processed: %s\n', filepath);
    end
    
    timerTicks('fin_ticker', ticker_info, 'Collating timings and scores');
    
    % average out the timings and the extra infos
    for idx = 1:length(timings_fields)
        avg_timings.(timings_fields{idx}) = ...
            avg_timings.(timings_fields{idx}) / length(segobj_filepaths);
        
    end
    for idx = 1:length(extra_info_fields)
        % average just the ~NaN values across the columns
        tmp = avg_stats.(extra_info_fields{idx});
        incl = ~isnan(tmp);
        tmp(~incl) = 0;
        avg_stats.(extra_info_fields{idx}) = sum(tmp, 1) ./ sum(incl, 1);
    end

    % replace avg_best_overlap by avg best overlap per object over all
    % images. Also add new covering score, which supposes all the the GTs 
    % were in the same image. Also compute recall (detection rate) scores.
    if ~isempty(all_extra_info.best_overlaps)
        % handle the case when no category strings are provided
        if isempty(catg_strs)
            N = max(all_extra_info.gt_category_id);
            catg_strs = ...
                arrayfun(@(i) num2str(i), 1:N, 'UniformOutput',false);
        end
        
        avg_stats.avg_best_overlap = mean(all_extra_info.best_overlaps);
        avg_stats.collated_covering = ...
            sum(all_extra_info.best_overlaps .* ...
                double(all_extra_info.gt_szs)) / sum(all_extra_info.gt_szs);
        avg_stats.avg_best_overlaps_sgraphs = ...
                            mean(all_extra_info.best_overlaps_sgraphs, 1);
        avg_stats.all_sgraph_gt_wins = ...
                            uint32(sum(all_extra_info.sgraph_gt_wins,1));
        avg_stats.avg_bb_overlap = mean(all_extra_info.bb_overlaps);
        avg_stats.num_gts = uint32(numel(all_extra_info.best_overlaps));
        avg_stats.num_ims = uint32(numel(unique(all_extra_info.im_no)));
        avg_stats.gt_categories = catg_strs;
        N = numel(avg_stats.gt_categories);
        avg_stats.category_avg_overlaps = ...
            accumarray(all_extra_info.gt_category_id(:), ...
                       all_extra_info.best_overlaps, [N, 1], @mean)';
        avg_stats.num_category_gts = ...
            uint32(accumarray(all_extra_info.gt_category_id(:), 1, [N, 1])');
        avg_stats.seg_recall = ...
            single(sum(bsxfun(@ge, all_extra_info.best_overlaps', ...
                              [0.5 0.7 0.9])) / double(avg_stats.num_gts));
        if ~isempty(all_extra_info.bb_overlaps)
            avg_stats.bb_recall = ...
                single(sum(bsxfun(@ge, all_extra_info.bb_overlaps', ...
                              [0.5 0.7 0.9])) / double(avg_stats.num_gts));
        else
            avg_stats.bb_recall = nan(1,3);
        end
        assert(sum(avg_stats.all_sgraph_gt_wins) == avg_stats.num_gts);
    end
    
    if ~isempty(output_dir)
        if ~isempty(timings_fields)
            plot_timing_bar_chart(avg_timings, avg_stats, 'avg', ...
                              output_dir, fi, 1, force_timing_scale);
        end
        
        % produce a scores image, if there were some quantitative results
        if ~isempty(all_extra_info.best_overlaps)
            scores_filepath = fullfile(output_dir, 'scores.fig');
            plot_score_graphs(avg_stats, all_extra_info, scores_filepath);
        end
        
        % save the average stats (also store the last seg_obj)
        if exist('load_mat','var') && isfield(load_mat, 'seg_obj')
            seg_obj = load_mat.seg_obj;
        else
            seg_obj = [];
        end
        save(fullfile(output_dir, 'avgstats.mat'), 'avg_timings', ...
                                                   'avg_stats', ...
                                                   'all_extra_info', ...
                                                   'seg_obj', ...
                                                   'segobj_filepaths');
    end

    dFig('close_fig', fi);
end


function plot_timing_bar_chart(all_timings, extra_info, filename, ...
                               output_dir, fi, draw_cmprsd, ...
                               force_timing_scale)
    
    if ~exist('force_timing_scale','var') || isempty(force_timing_scale)
        force_timing_scale = [];
    end
    
    cut_prct_time = all_timings.pmc_parallel_cut_time / ...
        (all_timings.pmc_parallel_cut_time + ...
         all_timings.pmc_parallel_overhead_time);
    
    timings_str = {'Boundary compute', ...
                   'Superpixel compute', ...
                   'Pairwise precompute', ...
                   'Seeds precompute', ...
                   'Pairwise init overhead', ...
                   'Unary potential compute', ...
                   'Pairwise potential compute', ...
                   'PMC time', ...
                   'PMC overhead time', ...
                   'Segment filtering init', ...
                   'Segment energy filter', ...
                   'Segment random filter', ...
                   'Segment similar filter', ...
                   'Graph setup/PMC/Filter overhead', ...
                   'All other overheads'};
    timings_plot = [all_timings.im_pairwise_time, ...
                    all_timings.superpixels_compute_time, ...
                    all_timings.precompute_pairwise_time, ...
                    all_timings.precompute_seed_time, ...
                    all_timings.total_init_time - ...
                        (all_timings.im_pairwise_time + ...
                         all_timings.superpixels_compute_time + ...
                         all_timings.precompute_pairwise_time + ...
                         all_timings.precompute_seed_time), ...
                    all_timings.unary_cost_set_time, ...
                    all_timings.pairwise_set_time, ...
                    all_timings.pmc_time * cut_prct_time, ...
                    all_timings.pmc_time * (1-cut_prct_time), ...
                    all_timings.init_filter_time ...
                    all_timings.energy_filter_time, ...
                    all_timings.rand_filter_time, ...
                    all_timings.seg_similar_filter_time, ...
                    all_timings.total_computing_segs_time - ...
                        (all_timings.unary_cost_set_time + ...
                         all_timings.pairwise_set_time + ...
                         all_timings.pmc_time + ...
                         all_timings.init_filter_time + ...
                         all_timings.energy_filter_time + ...
                         all_timings.rand_filter_time + ...
                         all_timings.seg_similar_filter_time), ...
                    all_timings.total_seg_time - ...
                        (all_timings.total_init_time + ...
                         all_timings.total_computing_segs_time)];
    color_ids = [1, 1, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3, 4, 5];
    clrs = [0.7 0.3 0; 0.2 0.2 0.8; 0.2 0.8 0.2; 0.2 0.6 0.6; 0.5 0.5 0.5];
    
    header_txts = {};
    
    header_txts{end+1} = sprintf('Total time: %.3fs', ...
                                all_timings.total_seg_time);
    if isfield(extra_info, 'avg_best_overlap')
        header_txts{end+1} = sprintf(['Total num segs.: %d\n', ...
                                  'Mean best overlap: %.4f\n', ...
                                  'Best covering: %.4f'], ...
                                 round(extra_info.final_num_segs), ...
                                 extra_info.avg_best_overlap, ...
                                 extra_info.covering);
    else
        header_txts{end+1} = sprintf(['Total num segs.: %d\n'], ...
                                 round(extra_info.final_num_segs));
    end
    header_font_szs = [12, 8];
    
%     out_filepath = fullfile(output_dir, [filename, '_timing.png']);
    out_filepath = fullfile(output_dir, [filename, '_timing.eps']);
    
    plot_print_bar(fi, timings_plot, timings_str, clrs, ...
                   color_ids, header_txts, header_font_szs, ...
                   out_filepath, force_timing_scale);
    
    if exist('draw_cmprsd','var') && draw_cmprsd == 1
        cmprsd_timings_str = {'Boundary compute', ...
                              'Superpixel compute', ...
                              'Pairwise compute', ...
                              'Unary compute', ...
                              'Parametric min-cut', ...
                              'Segment filteration', ...
                              'Overheads'};
        color_ids = [1, 1, 2, 2, 3, 4, 5];
    
        mapping = [1, 2, 3, 4, 3, 4, 3, 5, 5, 6, 6, 6, 6, 7, 7]';
        cmprsd_timings_plot = accumarray(mapping, timings_plot')';
        
        out_filepath = fullfile(output_dir, ...
                                [filename, '_timing_summary.eps']);
        
        plot_print_bar(fi, cmprsd_timings_plot, ...
                       cmprsd_timings_str, clrs, ...
                       color_ids, header_txts, header_font_szs, ...
                       out_filepath, force_timing_scale)
    end
end


function plot_print_bar(fi, timings_plot, timings_str, clrs, ...
                        color_ids, header_txts, header_font_szs, ...
                        filepath, force_timing_scale)
    fig_h = dFig('get_fig_h', fi);
    ax_h = dFig('get_ax_h', fi);
    
    N = length(timings_plot);
    
    % plot the bars
    for i = 1:N
        h = bar(ax_h, i, timings_plot(i));
        if i == 1, hold(ax_h, 'on'); end
        set(h, 'FaceColor', clrs(color_ids(i),:));
    end
    
    % switch off the Xtick text
    set(ax_h, 'XTickLabel',{});
    
    % set the ylims if instructed by the user
    if exist('force_timing_scale','var') && ~isempty(force_timing_scale)
        ylim(ax_h, force_timing_scale);
    end
    
    % xlabel rotated texts
    ypos = -max(ylim)/50;
    text(1:N, repmat(ypos,N,1), ...
         timings_str','HorizontalAlignment','Right','Rotation',90, ...
         'FontSize',10, 'Parent',ax_h);
    % ylabel text
    ylabel(ax_h, 'computation time (secs)','FontSize',12);
    
    % adjust the position of the graph to accomodate the vertical xlabels
    pos = get(ax_h, 'Position');
    set(ax_h, 'Position', [pos(1) 0.45 pos(3) 0.53]);
    
    extra_txt_pos = [max(xlim(ax_h)) - (max(xlim(ax_h))/20), ...
                     max(ylim(ax_h)) - (max(ylim(ax_h))/20)];
    
    % write different stats in the top right corner of the graph
    for h_idx = 1:length(header_txts)
        text(extra_txt_pos(1), extra_txt_pos(2) - ((h_idx-1) * max(ylim)/12), ...
             header_txts{h_idx}, 'HorizontalAlignment','Right', ...
             'VerticalAlignment','Top', 'FontSize', ...
             header_font_szs(h_idx), 'Parent',ax_h);
    end
    
%     export_fig(filepath, '-a1', '-transparent', fig_h);
%     saveas(fig_h, filepath);
    print(fig_h, filepath, '-depsc2');
    
    delete(get(ax_h, 'children'));
end


function plot_score_graphs(avg_stats, all_extra_info, filepath)
    fi = dFig('init_fig', [700, 1700], 'visible','off');
    
    recall_thrshs = 0:0.01:1;
    recall = sum(bsxfun(@gt, all_extra_info.best_overlaps', ...
                        recall_thrshs)) / double(avg_stats.num_gts);
    if ~isempty(all_extra_info.bb_overlaps)
        recall_bb = sum(bsxfun(@gt, all_extra_info.bb_overlaps', ...
                               recall_thrshs)) / double(avg_stats.num_gts);
    else
        recall_bb = zeros(1, numel(recall_thrshs));
    end
    subplot(dFig('get_ax_h', fi));
    
    % detection rate with different jaccard thresholds
    ax_h1 = subplot(2,3,1);
    plot(ax_h1, recall_thrshs, recall, 'LineWidth',2, 'Color',[0,0,1]);
    hold(ax_h1, 'on');
    plot(ax_h1, recall_thrshs, recall_bb, 'LineWidth',2, 'Color',[1,0,0]);
    legend({'Segmentation', 'Bounding Box'});
    xlabel('\(\mathcal{J}\)', 'Interpreter','latex');
    ylabel({'Detection rate / Recall';'% segs/BB above a certain overlap'}, ...
           'Interpreter','none');
    marker_pnts = 0.1:0.1:0.9;
    [~,idx] = min(abs(bsxfun(@minus, recall_thrshs', marker_pnts)));
    plot(marker_pnts, recall(idx), 'LineStyle','none', 'Marker','o', ...
         'MarkerEdgeColor',[0,0,1], 'MarkerFaceColor',[1,1,1]);
    for i = [5 7 9]
        x = marker_pnts(i);
        text(x-0.01, recall(idx(i))-0.01, sprintf('%.1f',x), ...
             'Color',[0 0 1], 'HorizontalAlignment','Right', ...
             'VerticalAlignment','Top');
        text(x+0.01, recall(idx(i))+0.01, ...
             sprintf('%.2f',avg_stats.seg_recall(i == [5 7 9])*100), ...
             'Color',[0 0 1], 'HorizontalAlignment','Left', ...
             'VerticalAlignment','Bottom');
    end
    % write down some per segmenter subgraph scores
    sgraph_txts = cell(numel(all_extra_info.sgraph_names), 1);
    for si = 1:numel(sgraph_txts)
        sgraph_txts{si} = sprintf('%s: %.2f (%d)', ...
                        all_extra_info.sgraph_names{si}, ...
                        avg_stats.avg_best_overlaps_sgraphs(si)*100, ...
                        avg_stats.all_sgraph_gt_wins(si));
    end
    text(0.05, 0.05, sgraph_txts, 'Parent',ax_h1, ...
         'VerticalAlignment','Bottom', 'HorizontalAlignment','Left');
    set(ax_h1, 'Position', [0.05 0.58 0.22 0.38]);
    
    % Overlap against GT sz
    NUM_SECTS = 20;
    ax_h2 = subplot(2,3,2);
    [gt_szs, sidxs] = sort(double(all_extra_info.gt_szs));
    % remove incoherent gt_szs
    sidxs(gt_szs <= 0) = [];  gt_szs(gt_szs <= 0) = [];
    overlaps = all_extra_info.best_overlaps(sidxs);
    sz_lims = logspace(log10(min(gt_szs)), log10(max(gt_szs)), NUM_SECTS*2);
    % find the starting index of each sz limit section
    first_idxs = arrayfun(@(l) find(gt_szs >= l, 1, 'first'), ...
                          sz_lims(1:2:end));
    first_idxs(1) = 1;  % force the first GT to be in the first section
    sz_sect = zeros(1, numel(gt_szs));
    for idx = first_idxs, sz_sect(idx) = sz_sect(idx) + 1; end
    sz_sect = cumsum(sz_sect);
    assert(sz_sect(end) == NUM_SECTS);
    % compute the min, max, mean overlap for each gt sz section
    valid_sects = unique(sz_sect);
    ov_mean = arrayfun(@(l) mean(overlaps(sz_sect == l)), valid_sects);
    ov_std = arrayfun(@(l) std(overlaps(sz_sect == l)), valid_sects);
    ov_max = arrayfun(@(l) max(overlaps(sz_sect == l)), valid_sects);
    ov_min = arrayfun(@(l) min(overlaps(sz_sect == l)), valid_sects);
    sz_marks = sz_lims(valid_sects*2);
    fill([sz_marks fliplr(sz_marks)], [ov_max fliplr(ov_min)], [0 1 1], ...
          'Parent',ax_h2);
    hold(ax_h2, 'on');
    fill([sz_marks fliplr(sz_marks)], ...
          [ov_mean+ov_std fliplr(ov_mean-ov_std)], [0 0.5 1], ...
          'Parent',ax_h2, 'FaceAlpha',0.5);
    plot(sz_marks, ov_mean, 'Parent',ax_h2, 'LineWidth',2, 'Color',[0 0 0]);
    set(ax_h2, 'XScale','log');
    xlabel('GT size (#pixels)', 'Interpreter','none');
    ylabel('Overlap', 'Interpreter','none');
    set(ax_h2, 'Position', [0.33 0.58 0.22 0.38]);
    ylim(ax_h2, [0 1]);
    xlims = xlim(ax_h2);
    fg_cut_str = sprintf('%.0f, ', avg_stats.frac_all_fg_cut*100);
    fg_cut_str = fg_cut_str(1:end-2);
    text(xlims(2)*0.9, 0.17, ...
         {sprintf('Avg. GT size: %.1f pixels', mean(gt_szs)), ...
          sprintf('Full cuts %%: %s', fg_cut_str)}, 'Parent',ax_h2, ...
         'VerticalAlignment','Top', 'HorizontalAlignment','Right');
    
    % Percentage segs/BB against overlap
    NUM_SECTS = 20;
    cmap = [linspace(1, 0, NUM_SECTS)' zeros(NUM_SECTS,1) ...
            linspace(0, 1, NUM_SECTS)'];
    ax_h4 = subplot(2,3,4);
    ax_h5 = subplot(2,3,5);
    xlim(ax_h4, [0, 1]);
    xlim(ax_h5, [0, 1]);
    bin_centers = 1/(NUM_SECTS*2):(1/NUM_SECTS):1;
    hist_overlap = hist(all_extra_info.best_overlaps, bin_centers);
    hist_bb_overlap = hist(all_extra_info.bb_overlaps, bin_centers);
    hold(ax_h4, 'on');
    hold(ax_h5, 'on');
    num_gts = double(avg_stats.num_gts);
    for idx = 1:length(hist_overlap)
        bar(bin_centers(idx), hist_overlap(idx)/num_gts, ...
            1/NUM_SECTS - 0.005, 'parent', ax_h4, 'facecolor', cmap(idx,:));
        bar(bin_centers(idx), hist_bb_overlap(idx)/num_gts, ...
            1/NUM_SECTS - 0.005, 'parent', ax_h5, 'facecolor', cmap(idx,:));
    end
    xlabel(ax_h4, 'Overlap', 'Interpreter','none');
    ylabel(ax_h4, '% segments', 'Interpreter','none');
    xlabel(ax_h5, 'Bounding Box Overlap', 'Interpreter','none');
    ylabel(ax_h5, '% bounding boxes', 'Interpreter','none');
    box(ax_h4,'on');
    box(ax_h5,'on');
    set(ax_h4, 'Position', [0.05 0.08 0.22 0.38]);
    set(ax_h5, 'Position', [0.33 0.08 0.22 0.38]);
    ylims1 = ylim(ax_h4);
    ylims2 = ylim(ax_h5);
    text(0.05, ylims1(2)*0.95, ...
         {sprintf('Avg. Overlap: %.2f', avg_stats.avg_best_overlap*100), ...
          sprintf('Covering: %.2f', avg_stats.covering*100), ...
          sprintf('%d GTs in %d ims',avg_stats.num_gts,avg_stats.num_ims), ...
          sprintf('Avg. # proposals: %.1f', avg_stats.final_num_segs)}, ...
         'Parent',ax_h4, 'VerticalAlignment','Top', ...
         'HorizontalAlignment','Left');
    text(0.05, ylims2(2)*0.95, ...
         {sprintf('Avg. BB Overlap: %.2f', avg_stats.avg_bb_overlap*100), ...
          sprintf('%d GTs in %d ims',avg_stats.num_gts,avg_stats.num_ims)}, ...
         'Parent',ax_h5, 'VerticalAlignment','Top', ...
         'HorizontalAlignment','Left');
    
    % per category overlaps
    MAX_CAT_PER_GRAPH = 50;
    per_cat_overlap = avg_stats.category_avg_overlaps;
    num_cat_gt = avg_stats.num_category_gts;
    num_catg = numel(avg_stats.gt_categories);
    assert(numel(per_cat_overlap) <= num_catg, ...
           'Number of categories is more than the dataset specifies');
    idx = 1;
    while true
        ax_h = subplot(2,3,idx*3);
        
        % get data for this plot
        start_idx = (idx-1)*MAX_CAT_PER_GRAPH + 1;
        end_idx = min(num_catg, idx*MAX_CAT_PER_GRAPH);
        curr_overlaps = per_cat_overlap(start_idx:end_idx);
        curr_num_gts = num_cat_gt(start_idx:end_idx);
        curr_strs = avg_stats.gt_categories(start_idx:end_idx);
        bar(curr_overlaps, 'FaceColor', [1 1 0], 'BarWidth',0.7);
        
        N = numel(curr_num_gts);
        xlim([0 N+1]);
        ylim([0 1]);
        
        % switch off xticklabels
        set(ax_h,'XTick',[]);
        ypos = -1/50;
        % write the vertical xlabel texts
        text(1:N, repmat(ypos,N,1), curr_strs', ...
             'HorizontalAlignment','Right', 'VerticalAlignment','Middle', ...
             'Rotation',90, 'FontSize',10, 'Parent',ax_h);
        curr_num_gts = cellfun(@(x) num2str(x), ...
                        num2cell(curr_num_gts), 'UniformOutput',false);
        text(1:N, repmat(-ypos,N,1), curr_num_gts', ...
             'HorizontalAlignment','Left', 'VerticalAlignment','Middle', ...
             'Rotation',90, 'FontSize',10, 'Parent',ax_h);
        % ylabel text
        ylabel(ax_h, 'Overlap','FontSize',12);
        
        % stretch the graph horizontally
        if idx == 1
            set(ax_h, 'Position', [0.61 0.62 0.36 0.34]);
        else
            set(ax_h, 'Position', [0.61 0.12 0.36 0.34]);
        end
        
        idx = idx + 1;
        if idx > 2 || end_idx >= num_catg
            break;
        end  
    end
    
    set(dFig('get_fig_h', fi), 'Visible','on'); 
    saveas(dFig('get_fig_h', fi), filepath);
%     print(dFig('get_fig_h', fi), filepath, '-depsc2');
    dFig('close_fig', fi);
end
