function run_user_seed_rigor()
% Test function demonstrating how seeds can be provided manually to
% rigor_obj_segments. In this function the seed is defined by mouse click
% on an image.
%
% @authors:     Ahmad Humayun
% @contact:     ahumayun@cc.gatech.edu
% @affiliation: Georgia Institute of Technology
% @date:        Fall 2013 - Summer 2014

% add RIGOR and utils to the path
rigor_root = fullfile(fileparts(which(mfilename)), '..');
addpath(rigor_root);
addpath(fullfile(rigor_root, 'utils'));

im_filepath = 'peppers.png';
testImg = imread(im_filepath);
szim = size(testImg);
fi = dFig('init_fig', szim);
dFig('imshow_fig', testImg, fi);
set(gcf, 'Position', [300 300 size(testImg,2)*1.3 size(testImg,1)*1.3]);
fprintf('Please select a point on the image as a seed\n')
seed_centers = ginput(1);
seed_centers = [seed_centers(2) seed_centers(1)];
seed_centers = max(seed_centers, 1);
seed_centers = min(seed_centers, szim(1:2));

% each row of this matrix gives a user specified seed location 
% ([row, column] format)
% seed_centers = [100 200;   300 500;   200 1];
% seed_centers = [100 200];

% this is the size of the seed region around each selected point. a [1 1] 
% just means a single pixel
seed_sz = [50 50];

% run RIGOR with user specified seeds
[segs, sp_seg, seg_obj, total_time] = ...
    rigor_obj_segments(im_filepath, ...
                      'graph_sub_methods_seeds_idx', {[1, 1, 1], [1, 1, 1]}, ...
                      'graph_seed_gen_method', {'gen_user_seeds'}, ...
                      'graph_seed_nums', {seed_sz}, ...
                      'graph_seed_params', {{seed_centers}}, ...
                      'force_recompute',true);

% converting superpixel segments into full image segments
segs = convert_masks(sp_seg, segs);

% display settings
dilate_bndr_sz = 1;
alpha_bndr = 1;
alpha_val = 0.6;

% get some statistics to find how each segment was generated
segs_per_graph = seg_obj.num_segs.after_clustering_FINAL;
temp_graph_cnt = cumsum([0 segs_per_graph]);
I = imread(im_filepath);
sz = size(I);
alpha_mat = ones(sz([1 2])) * alpha_val;
alpha_bndr_mat = ones(sz([1 2])) * alpha_bndr;
seg_clrs = jet(size(segs,3));

fi2 = dFig('init_fig', size(sp_seg));

%% iterate over results and display
for idx = 1:size(segs,3)
    % get some more info about the current segment
    graph_type_idx = find(idx >= cumsum([1 segs_per_graph]), 1, 'last');
    seg_idx = idx - temp_graph_cnt(graph_type_idx);
    sol_idx = ...
        seg_obj.meta_cut_segs_info(graph_type_idx).seg_mapping_final_to_orig(seg_idx);
    lambda = seg_obj.meta_cut_segs_info(graph_type_idx).lambdas(sol_idx);
    subgraph_idx = ...
        seg_obj.meta_cut_segs_info(graph_type_idx).sols_to_unary_mapping(sol_idx);
    graph_method = seg_obj.segm_params.graph_methods{graph_type_idx};
    subgraph_method = ...
        seg_obj.segm_params.graph_sub_methods{graph_type_idx}{subgraph_idx};
    
    % draw the input image
    dFig('imshow_fig', I, fi2);
    dFig('holdonoff','on', fi2);
    
    % draw the seg with some transparency
    curr_alpha_mat = alpha_mat;
    curr_alpha_mat(~segs(:,:,idx)) = 0;
    seg_im = bsxfun(@times, im2double(segs(:,:,idx)), ...
                            reshape(seg_clrs(idx,:), [1 1 3]));
    dFig('image_fig',seg_im, fi2, 'AlphaData',curr_alpha_mat);
                 
    % draw boundary around the seg
    bndr = bwperim(segs(:,:,idx), 4);
    bndr = imdilate(bndr, strel('disk', dilate_bndr_sz));
    draw_bndr = im2double(repmat(bndr, [1 1 3]));
    curr_alpha_bndr_mat = alpha_bndr_mat;
    curr_alpha_bndr_mat(~bndr) = 0;
    dFig('image_fig',draw_bndr, fi2, 'AlphaData',curr_alpha_bndr_mat);
    
    % draw the point click (seed)
    plot(seed_centers(:,2), seed_centers(:,1), 'rx', 'MarkerSize', 10)
    
    % display the extra information about the seg displayed
    txt = sprintf('Segment %d out of %d (%s %s lambda %.3f)', idx, ...
                  size(segs,3), graph_method, subgraph_method, lambda);
    dFig('title_fig',txt, fi2, 'Interpreter','latex', 'FontSize',14);
    
    %dispMask(testImg, segs(:,:,idx), 0.1);
    pause;
end

dFig('close_fig', fi2);
