/* 
 * @authors:     Fuxin Li
 * @contact:     fli@cc.gatech.edu
 * @affiliation: Georgia Institute of Technology
 * @date:        Fall 2013 - Fall 2014
 */

#include "mex.h"
#include "stdint.h"
#include "math.h"
#include <vector>

using namespace std;

bool isadjacent(uint32_t x1,uint32_t y1, uint32_t x2, uint32_t y2)
{
	// uint, cannot directly do minus
	if((x1 >= x2 && x1 - x2 <= 1) || (x2 > x1 && x2 - x1 <= 1))
		if((y1 >= y2 && y1 - y2 <= 1) || (y2 > y1 && y2 - y1 <= 1))
			return true;
	return false;
}

bool check_line(double eigx, double eigy, const vector<uint32_t> &points, const vector<uint32_t> &inds, const uint32_t *locx, const uint32_t *locy)
{
	uint32_t i;
	double sumx, sumy, sumxx, sumyy, sumxy;
	double allb1, allb2;
	double this_eigx, this_eigy;
	sumx = sumy = sumxx = sumyy = sumxy = 0;
	for(i=0;i<inds.size();i++)
	{
		uint32_t the_ind = points[inds[i]];
		sumx += locx[the_ind];
		sumy += locy[the_ind];
		sumxx += locx[the_ind] * locx[the_ind];
		sumyy += locy[the_ind] * locy[the_ind];
		sumxy += locx[the_ind] * locy[the_ind];
	}
	sumx /= inds.size();
	sumy /= inds.size();
	sumxx /= inds.size();
	sumyy /= inds.size();
	sumxy /= inds.size();
        sumxy -= sumx * sumy;
        sumxx -= sumx * sumx;
        sumyy -= sumy * sumy;
        double the_tr = sumxx + sumyy;
        double the_det = sumxx * sumyy - sumxy * sumxy;
        double eig1 = the_tr /2.0 + sqrt((the_tr*the_tr/4.0 - the_det));
	// Divide by 0
	if (fabs(sumxy) < 1e-5 && sumyy > sumxx)
		this_eigx = 0.0, this_eigy = 1.0;
	else if (fabs(sumxy) < 1e-5 && sumxx >= sumyy)
		this_eigx = 1.0, this_eigy = 0.0;
	else
	{
		double sumsq = sqrt((eig1 - sumyy) * (eig1 - sumyy) + sumxy * sumxy);
		this_eigx = (eig1 - sumyy) / sumsq;
		this_eigy = sumxy / sumsq;
	}
	if (fabs(eigx * this_eigx + eigy * this_eigy) >= 0.85)
		return true;
	return false;
}

void filter_points_major_dir(const uint32_t *vals, const double *eigx, const double *eigy, const uint32_t *locx, const uint32_t *locy,
	                         const uint32_t num_points, const uint32_t num_sups, bool *locsind)
{
	vector<uint32_t> *all_sups = new vector<uint32_t>[num_sups*num_sups];
	uint32_t i,j,k,ii;
	for(i=0;i<num_points;i++)
	{
		// Not doing stuff when x position is 0
		if (locx[i] == 0)
			continue;
		// For this round, just push everything into the corresponding all_sups
		for(j=0;j<8;j++)
		{
			if (vals[i * 8 + j] == 0)
				break;
			for(k=j+1;k<8;k++)
			{
				if (vals[i*8+k] == 0)
					break;
				all_sups[num_sups * (vals[i *8 + j] - 1) + vals[i * 8 + k] - 1].push_back(i);
			}
		}
	}
	// Second round: determine connected components and the eigenvectors of all connected components
	uint32_t idx = 0;
	for(i=0;i<num_sups;i++)
	{
		idx = i * num_sups + i + 1;
		for (j=i+1;j<num_sups;j++,idx++)
		{
			// Many vectors will be empty, doesn't matter
			if (all_sups[idx].empty())
				continue;
			// If not empty, determine the connected components here, just use stupid floodfill is OK.
			bool change_idx = true;
			bool *done_flag = new bool[all_sups[idx].size()]();
			uint32_t min_free_idx = 0;
			while(min_free_idx < all_sups[idx].size())
			{
			// Initialize
				vector<uint32_t> inside_idx;
				done_flag[min_free_idx] = true;
				inside_idx.push_back(min_free_idx);
				// Floodfill, run stupid algorithm without a priority queue, because we don't care about the speed of this
				for(k=0;k<inside_idx.size();k++)
				{
					for(ii=min_free_idx+1;ii<all_sups[idx].size();ii++)
					{
						if (done_flag[ii])
							continue;
						if (isadjacent(locx[all_sups[idx][inside_idx[k]]],locy[all_sups[idx][inside_idx[k]]],
							          locx[all_sups[idx][ii]], locy[all_sups[idx][ii]]))
						{
							done_flag[ii] = true;
							inside_idx.push_back(ii);
						}
					}
				}
				bool match = check_line(eigx[j*num_sups+i], eigy[j*num_sups+i], all_sups[idx], inside_idx, locx, locy);
				if (match)
				{
					for (k=0;k<inside_idx.size();k++)
						locsind[all_sups[idx][inside_idx[k]]] = true;
				}
				for(ii=min_free_idx+1;ii<all_sups[idx].size();ii++)
				{
					if (!done_flag[ii])
						break;
				}
				min_free_idx = ii;
			}
			delete []done_flag;
		}
	}
	delete []all_sups;
}

void mexFunction( int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[] )
{
	if (nrhs < 5)
		mexErrMsgTxt("Need five input arguments. Values (uint32) , direction_x (double), direction_y (double), matched points x and y (uint32)");
	uint32_t *vals = (uint32_t *) mxGetData(prhs[0]);
	double *eigx = mxGetPr(prhs[1]);
	double *eigy = mxGetPr(prhs[2]);
	uint32_t *locx = (uint32_t *) mxGetData(prhs[3]);
	uint32_t *locy = (uint32_t *) mxGetData(prhs[4]);
	uint32_t num_points, num_sups;

	if (vals == 0 || eigx == 0 || eigy == 0 || locx == 0 || locy == 0)
		mexErrMsgTxt("Can't read one of the arguments! Please check the 5 arguments!");

	num_points = mxGetN(prhs[0]);
	num_sups = mxGetM(prhs[1]);

	plhs[0] = mxCreateLogicalMatrix(num_points, 1);
	bool *locsind =  (bool *) mxGetData(plhs[0]);

	filter_points_major_dir(vals, eigx, eigy, locx, locy, num_points, num_sups, locsind);
}
