function compute_print_results(segs, seg_obj, gt_fldr, dataset_id)
% You can use this function to compute accuracy scores for the segment 
% proposals generated against some GT. This function computes the overlap 
% for each segment to all GTs. Then prints the results and stores them. The 
% results are stored at the path defined by scores_dirpath defined in 
% internal_params.m. The file stores some collated scores including 
% overlap, and covering (collated_scores). It also stores the number of 
% segments generated by parametric min-cut, and the number remaining after 
% filtration (init_num_segs and final_num_segs). Also overlaps of all 
% segments against the GT is stored (Q). The function finally displays the 
% results in a pretty table :)
%
% For instance, if you are working with PASCAL VOC (e.g. stored at 
%  ~/pascalvoc12), you can do the following:
%       [segs, sp_seg, seg_obj, total_time] = ...
%           rigor_obj_segments('~/pascalvoc12/JPEGImages/2007_000033.jpg');
%       segs = convert_masks(sp_seg, segs);  % convert to binary masks
%       compute_print_results(segs, seg_obj, ...
%           '~/pascalvoc12/SegmentationObject');
%
% You can also run compute_print_results.m by loading the data saved to 
% disk by rigor_obj_segments.m, and supplying the converted binary mask 
% segs, and seg_obj to the function.
%
% @authors:     Ahmad Humayun
% @contact:     ahumayun@cc.gatech.edu
% @affiliation: Georgia Institute of Technology
% @date:        Fall 2013 - Summer 2014

    if ~exist('dataset_id', 'var'), dataset_id = 'pascalvoc'; end
    
    assert(size(segs,3) == sum(seg_obj.num_segs.after_clustering_FINAL));
    
    img_name = seg_obj.input_info.img_name;
    
    % compute the quality of each segment given the GT
    [Q, collated_scores] = proposal_quality(img_name, gt_fldr, ...
                                segs, 'overlap', dataset_id, ...
                                seg_obj.filepath_params.extern_src_dir);
    
    % get individual scores for each segmenter and its sub-graphs
    [collated_scores, einfo] = split_scores(Q, seg_obj, collated_scores);
    
    % print the table of scores
    print_scores_table([Q.gt_seg_szs], einfo, collated_scores);
    
    % save the scores
    scores_filepath = seg_obj.return_scores_filepath();
    
    dir_name = fileparts(scores_filepath);
    if(~exist(dir_name, 'dir'))
        mkdir(dir_name);
    end
    save(scores_filepath, 'Q', 'einfo', 'collated_scores');

    % print a summary line, useful for noting down results/performance
    print_summary(seg_obj, collated_scores);
end


function print_summary(seg_obj, cscores)
    for idx = 1:numel(seg_obj.segm_params.graph_methods)
        fprintf(1, '% 4d (% 4d)', ...
                seg_obj.num_segs.init_segs(idx), ...
                seg_obj.num_segs.after_clustering_FINAL(idx));
        if idx < numel(seg_obj.segm_params.graph_methods)
            fprintf(1, ', ');
        else
            fprintf(1, ' - ');
        end
    end
    best_gt_scores = max(cscores.best_seg_overlap_segmenters,[],2);
    for idx = 1:length(best_gt_scores)
        fprintf(1, '%.4f', best_gt_scores(idx));
        if idx < length(best_gt_scores)
            fprintf(1, ' + ');
        else
            fprintf(1, ' => ');
        end
    end
    for idx = 1:length(cscores.avg_best_overlap_segmenters)
        fprintf(1, '%.4f', cscores.avg_best_overlap_segmenters(idx));
        if idx < length(cscores.avg_best_overlap_segmenters)
            fprintf(1, ' + ');
        else
            fprintf(1, ' => ');
        end
    end
    fprintf('%.4f\n', cscores.avg_best_overlap);
end
