function print_scores_table(gt_seg_szs, einfo, cscores)
% Prints different scores in a tabular form. Given the individual overlap
% scores for each sub-graph, this function prints those scores in a table
% against each GT. The function also prints the number of segments produced
% by each sub-graph.
%
% @authors:     Ahmad Humayun
% @contact:     ahumayun@cc.gatech.edu
% @affiliation: Georgia Institute of Technology
% @date:        Fall 2013 - Spring 2015

    % get the row headers
    if ~isempty(gt_seg_szs)
        row_txts = arrayfun(@(i,gt_sz) sprintf('GT %2d (%5d)', i, gt_sz), ...
                            1:length(gt_seg_szs), gt_seg_szs, ...
                            'UniformOutput', false);
    else
        row_txts = {};
    end
    row_txts = [{'', 'Init. # Segs', 'Final # Segs', '', 'Avg. Best Overlap', ...
                'Covering'}, row_txts];
    
    num_subgraphs = accumarray(cscores.subgraph_to_sgmntr_idx(:), 1)';
    
    % get the column headers
    col_txts = [{''}, einfo.sgmntr_names, {'Overall'}];
    sgraph_shrt = cellfun(@(x) x([1:3 end]), einfo.sgraph_names, ...
                          'UniformOutput',false);
    col_txts2 = [{''}, sgraph_shrt, {''}];
    
    % compute the width based on the header lengths
    first_col_width = max(cellfun(@length, row_txts)) + 1;
    other_col_width = max(cellfun(@length, [sgraph_shrt,col_txts(end)])) + 1;
    
    % create a divider line
    total_width = 2 + first_col_width + ...
                  (length(col_txts2) - 1)*(other_col_width + 1);
    divider_line = [repmat('-', 1, total_width), '\n'];
    
    fprintf(divider_line);
    
    lastidx = length(einfo.sgmntr_names)+1;
    lastidx2 = length(einfo.sgraph_names)+1;
    
    % printf headers
    print_line(row_txts{1}, col_txts(2:end), 's', first_col_width, ...
               other_col_width, [num_subgraphs 1], lastidx);
    print_line(row_txts{1}, col_txts2(2:end), 's', first_col_width, ...
               other_col_width);
    
    fprintf(divider_line);
    
    % print initial number of segments
    print_line(row_txts{2}, ...
               [einfo.init_num_segs_sgraphs, ...
                sum(einfo.init_num_segs_sgraphs)], 'd', ...
               first_col_width, other_col_width);
    
    % print final number of segments
    print_line(row_txts{3}, einfo.final_num_segs_sgraphs, 'd', ...
               first_col_width, other_col_width, [], [], 1);
    print_line(row_txts{4}, ...
               [einfo.final_num_segs, sum(einfo.final_num_segs)], 'd', ...
               first_col_width, other_col_width, [num_subgraphs 1], ...
               lastidx);
           
    fprintf(divider_line);
    
    % print average best overlap
    [~, best_idx] = max(cscores.avg_best_overlap_sgraphs);
    print_line(row_txts{5}, [cscores.avg_best_overlap_sgraphs, ...
                             cscores.avg_best_overlap], '.4f', ...
               first_col_width, other_col_width, [], [best_idx, lastidx2]);
    
    % print covering
    [~, best_idx] = max(cscores.covering_sgraphs);
    print_line(row_txts{6}, [cscores.covering_sgraphs, ...
                             cscores.covering], '.4f', ...
               first_col_width, other_col_width, [], best_idx);
    
    fprintf(divider_line);
    
    % print stats for each GT
    for idx = 1:size(cscores.best_seg_overlap_segmenters,1)
        [best_val, best_idx] = ...
            max(cscores.best_seg_overlap_sgraphs(idx,:));
        print_line(row_txts{6+idx}, ...
                   [cscores.best_seg_overlap_sgraphs(idx,:), ...
                    best_val], '.4f', first_col_width, other_col_width, ...
                   [], best_idx);
    end
    
    fprintf(divider_line);
end


function print_line(header, vals, spec, first_col_width, ...
                    other_col_width, varargin)
    % controls whether a certain value is going to be bold
    num_cols = ones(1, numel(vals));
    strong = false(1, length(vals));
    num_extra_cols = 0;
    
    % num_cols, strong_idx, num_extra_cols
    if numel(varargin) > 0
        if ~isempty(varargin{1}), num_cols = varargin{1}; end
    end
    if numel(varargin) > 1
        strong(varargin{2}) = true;
    end
    if numel(varargin) > 2
        num_extra_cols = varargin{3};
    end
    
    strong_o = repmat({'<strong>'}, 1, length(vals));
    strong_e = repmat({'</strong>'}, 1, length(vals));
    strong_o(~strong) = {''};
    strong_e(~strong) = {''};
    
    for idx = 1:length(vals)
        room = num_cols(idx)*(other_col_width+1)-1;
        if ~iscell(vals)
            l = length(num2str(vals(idx)));
        else
            l = length(vals{idx});
        end
        neg_spc = max(l, other_col_width);
        prefix = floor((room - neg_spc) / 2);
        postfix = ceil((room - neg_spc) / 2);
        if prefix > 0
            strong_o{idx} = [repmat(' ', 1, prefix), strong_o{idx}];
        end
        if postfix > 0
            strong_e{idx} = [strong_e{idx}, repmat(' ', 1, postfix)];
        end
    end
    
    % print header
    eval(['fprintf(''|%-', num2str(first_col_width), 's'', header)']);
    
    if spec == 's'
        % in case of string
        temp = [strong_o; vals; strong_e];
    else
        temp = [strong_o; num2cell(vals); strong_e];
    end
    eval(['fprintf(''|%s%-', num2str(other_col_width), spec '%s'', temp{:})']);
    
    % put in extra cols
    one_extra_col = ['|' repmat(' ', 1, other_col_width)];
    fprintf('%s', repmat(one_extra_col, 1, num_extra_cols));
    fprintf('|\n');
end
