function [ collated_scores, einfo ] = ...
                        split_scores(Q, seg_obj, collated_scores)
%SPLIT_SCORES computes scores and statistics for each individual segmenter 
% and its sub-graphs over all the GTs
%
% @authors:     Ahmad Humayun
% @contact:     ahumayun@cc.gatech.edu
% @affiliation: Georgia Institute of Technology
% @date:        Fall 2013 - Fall 2014

    % collect scores for all segments
    overlap_scores = [Q.q];
    gt_seg_szs = [Q.gt_seg_szs];
    % if there were no GTs
    if isempty(overlap_scores)
        num_segs = sum(seg_obj.num_segs.after_clustering_FINAL);
        overlap_scores = zeros(num_segs, 0, 'single');
        gt_seg_szs = zeros(1, 0, 'uint32');
    end
    
    num_gt = numel(Q);
    num_sgmntrs = numel(seg_obj.segm_params.graph_methods);
    num_subgraphs = cellfun(@numel, ...
        seg_obj.segm_params.graph_sub_methods(1:num_sgmntrs));
    tot_subgraphs = sum(num_subgraphs);
    
    % # of segments returned by each segmenter and the remaining number
    % after the filteration step
    einfo.init_num_segs = seg_obj.num_segs.init_segs;
    einfo.final_num_segs = seg_obj.num_segs.after_clustering_FINAL;
    
    % get the name of each segmenter and sub-graphs
    sgmntr_names = seg_obj.segm_params.graph_methods;
    einfo.sgmntr_names = cellfun(@(s) strrep(s,'Graph',''), ...
                                      sgmntr_names, 'UniformOutput',false);
    einfo.sgraph_names = ...
        horzcat(seg_obj.segm_params.graph_sub_methods{1:num_sgmntrs});
    
    % compute the scores over different segmenters individually
    start_idx = 1;
    subgraph_idx = 1;
    
    % init all collated score variables
    avg_best_overlap_segmenters = nan(1, num_sgmntrs, 'single');
    covering_segmenters = nan(1, num_sgmntrs, 'single');
    best_seg_overlap_segmenters = zeros(num_gt, num_sgmntrs, 'single');
    best_seg_idx_segmenters = repmat(int32(-1), num_gt, num_sgmntrs);
    init_num_segs_sgraphs = zeros(1, tot_subgraphs, 'uint16');
    final_num_segs_sgraphs = zeros(1, tot_subgraphs, 'uint16');
    avg_best_overlap_sgraphs = zeros(1, tot_subgraphs, 'single');
    covering_sgraphs = zeros(1, tot_subgraphs, 'single');
    best_seg_overlap_sgraphs = zeros(num_gt, tot_subgraphs, 'single');
    best_seg_idx_sgraphs = repmat(int32(-1), num_gt, tot_subgraphs);
    subgraph_to_sgmntr_idx = zeros(1, tot_subgraphs, 'uint8');
    
    % iterate over each segmenter
    for sgmntr_idx = 1:num_sgmntrs
        cut_info = seg_obj.meta_cut_segs_info(sgmntr_idx);
        cinfo = cut_info.extra_cut_info;
        curr_num = num_subgraphs(sgmntr_idx);
        
        s = subgraph_idx;
        e = subgraph_idx + curr_num - 1;    
        r = s:e;
        subgraph_to_sgmntr_idx(r) = sgmntr_idx;
        
        % get initial number of segments for each subgraph
        init_cut_graph_types = ...
            cinfo.cuts_graph_type_idxs(cinfo.seg_to_cut_mapping);
        curr_init_num_segs = ...
            accumarray(init_cut_graph_types(:), 1, [curr_num 1]);
        assert(sum(curr_init_num_segs) == einfo.init_num_segs(sgmntr_idx));
        init_num_segs_sgraphs(r) = curr_init_num_segs;
        
        % get final number of segments post filtration for each subgraph
        final_seg_to_cut_mapping = ...
            cinfo.seg_to_cut_mapping(cut_info.seg_mapping_final_to_orig);
        final_cut_graph_types = ...
            cinfo.cuts_graph_type_idxs(final_seg_to_cut_mapping);
        curr_final_num_segs = ...
            accumarray(final_cut_graph_types(:), 1, [curr_num 1]);
        assert(sum(curr_final_num_segs) == ...
               einfo.final_num_segs(sgmntr_idx));
        final_num_segs_sgraphs(r) = curr_final_num_segs;
        
        % get the scores for just the current segmenter (Ns x Gt)
        segmenter_overlap = overlap_scores(start_idx:start_idx + ...
                                    einfo.final_num_segs(sgmntr_idx)-1,:);
        % iterate over each subgraph to collect their best overlap scores
        for idx = 1:curr_num
            if curr_final_num_segs(idx) == 0
                continue;
            end
            curr_idxs = final_cut_graph_types == idx;
            [best_overlap, best_idx] = ...
                max(segmenter_overlap(curr_idxs,:), [], 1);
            tmp = find(curr_idxs);
            best_idx = uint32(tmp(best_idx));
            best_seg_overlap_sgraphs(:,r(idx)) = best_overlap;
            best_seg_idx_sgraphs(:,r(idx)) = start_idx + best_idx - 1;
        end
        avg_best_overlap_sgraphs(:,r) = ...
            mean(best_seg_overlap_sgraphs(:,r), 1);
        % compute the collective covering for all GT's for each of the 
        % current set of subgraphs
        covering_sgraphs(r) = ...
            sum(bsxfun(@times, best_seg_overlap_sgraphs(:,r), ...
                       double(gt_seg_szs)'), 1) / sum(gt_seg_szs);
        
        % compute the best scores for each GT for the current segmenter
        best_seg_overlap = zeros(1, num_gt, 'single');
        if ~isempty(segmenter_overlap)
            [best_seg_overlap, best_seg_idx] = max(segmenter_overlap,[],1);
            avg_best_overlap_segmenters(sgmntr_idx) = mean(best_seg_overlap);
            best_seg_overlap_segmenters(:,sgmntr_idx) = best_seg_overlap;
            best_seg_idx_segmenters(:,sgmntr_idx) = ...
                                    start_idx + uint32(best_seg_idx) - 1;
        end
         
        % compute the collective covering for all GT's for the current
        % segmenter
        covering_segmenters(sgmntr_idx) = ...
            sum(best_seg_overlap .* double(gt_seg_szs)) / sum(gt_seg_szs);

        % verify individual subgraph scores with overall segmenter scores
        nz_sg = r(curr_final_num_segs > 0);
        if any(nz_sg)    % if there are non-zero # of final segments
            [tmpv, tmpi] = max(best_seg_overlap_sgraphs(:,nz_sg), [], 2);
            assert(all(best_seg_overlap == tmpv'));
            tmp = best_seg_idx_sgraphs(:,nz_sg);
            tmp = tmp(sub2ind(size(tmp), 1:num_gt, tmpi'));
            assert(all(tmp(:) == best_seg_idx_segmenters(:,sgmntr_idx)));
        else
            assert(all(all(best_seg_overlap_sgraphs(:,r) == 0)));
            assert(all(best_seg_overlap == 0));
            assert(all(all(best_seg_idx_sgraphs(:,r) == -1)));
            assert(all(best_seg_idx_segmenters(:,sgmntr_idx)== -1));
        end
        
        % increment the indices
        start_idx = start_idx + einfo.final_num_segs(sgmntr_idx);
        subgraph_idx = subgraph_idx + curr_num;
    end
    
    % store the scores computed
    collated_scores.covering_segmenters = covering_segmenters;
    collated_scores.covering_sgraphs = covering_sgraphs;
    collated_scores.avg_best_overlap_segmenters = ...
                                            avg_best_overlap_segmenters;
    collated_scores.avg_best_overlap_sgraphs = ...
                                            avg_best_overlap_sgraphs;
    collated_scores.best_seg_overlap_segmenters = ...
                                            best_seg_overlap_segmenters;
    collated_scores.best_seg_overlap_sgraphs = best_seg_overlap_sgraphs;
    collated_scores.best_seg_idx_segmenters = best_seg_idx_segmenters;
    collated_scores.best_seg_idx_sgraphs = best_seg_idx_sgraphs;
    collated_scores.subgraph_to_sgmntr_idx = subgraph_to_sgmntr_idx;
    
    % append to extra info
    einfo.init_num_segs_sgraphs = init_num_segs_sgraphs;
    einfo.final_num_segs_sgraphs = final_num_segs_sgraphs;
end

